
/*	Project Reversi II
 * 	august 2002
 * 
 * 	Dirk Richter
 * 	Institute of Computer Sience
 * 	Marthin-Luther-University of Halle
 * 
 * 	Dies ist die grafische Aufarbeitung von Reversi I
 */


import java.util.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;


class Pos {
	Pos(int X, int Y) 	{ x=(byte)X; y=(byte)Y;}
	Pos(Pos p)		{ x=p.x; y=p.y; }
	public byte x,y;
}

class Zug implements Comparable {
	private static Random rnd = new Random();
	private float Punkte;
	private Pos pos;			// Wurzel soll besten Zug enthalten
	private static ArrayList subTree;	// ein Blatt enthllt hier ein Brett,  ein innerer Knoten eine ArrayList
	private static Brett intern_brett;

	public Zug(int x, int y)	{ pos=new Pos(x,y); subTree=null; }
	public Pos getPos() 		{ return new Pos(pos); }
	public float getPunkte()	{ return Punkte; }

	public boolean equals(Object obj) {		// gleich Zug gdw. gleiche Position
		Pos p = ((Zug)obj).getPos();
		return (pos.x==p.x)&&(pos.y==p.y);
	}
	
	public int compareTo(Object b) {			// Vergleich erfolgt mittels Punkte
		Float i1 = new Float(getPunkte());
		Float i2 = new Float(((Zug)b).getPunkte());
		return i2.compareTo(i1);
	}
	
	public Zug getSubTree(int nr) {
		if (subTree!=null) return (Zug)(subTree.get(nr));
		return null;
	}
	
	public Brett verwerfeBrett() {
		Brett res=null;
		res = intern_brett;
		subTree=null;
		intern_brett=null;
		return res;
	}
	
	private float BrettLaden(Brett brett) {	// erhlt orginal Brett, soll aus dem Zug neues Brett bauen und liefert die Punkte
		Brett kopie = new Brett(brett);	// kopieren
		kopie.moveTo(pos.x,pos.y);		// Zug durchfhren
		intern_brett = kopie;			// Brett speichern
		Punkte = kopie.BrettWert();		// berechnet den Wert des Bretts
		return Punkte;
	}
	
	public static String translate(int x, int y) {
		final byte code[]={(byte)('A'+y),(byte)('1'+x)};
		return new String(code);
	}

	public Pos getBestPos() {	// liefert per Zufall eine der Besten Positionen
		float p = ((Zug)subTree.get(0)).getPunkte();
		
		int anz=1;
		while (anz<subTree.size()) if (((Zug)subTree.get(anz)).getPunkte()==p) anz++; else break;
		
		return ((Zug)subTree.get(Math.abs(rnd.nextInt())%anz)).getPos();
	}

	public int Aufbauen(byte Tiefe, Brett brett) {	// liefert Anz. berechneter Zge
		/* Baumeigenschaften:
		 * 	- jeder innere Knoten (auch Wurzel) enthllt ArrayList(Zug) sortiert nach Punkte
		 * 	- jedes Blatt enthllt aktuelles Reversi-Brett
		 * 	- Punkte im Blatt
		 * 
		 * */

		// alle mglichen Zge ermitteln
		ArrayList moeglich = new ArrayList();
		for (int i=0; i<8; i++) for (int j=0; j<8; j++) if (brett.erlaubt(i,j)) moeglich.add(new Zug(i,j));
		moeglich.trimToSize();
		final byte len = (byte)(moeglich.size());
		int res = 0;

		Punkte = 0;		// auf Max setzen
		float min = Float.MAX_VALUE;
		
		if (0==len) {	// keine Freien Zge
			if (Tiefe<=1) {
				Punkte = brett.BrettWert();
				return 1;
			} else {
				Brett kopie = new Brett(brett);
				kopie.next_player();
				res = Aufbauen((byte)(Tiefe-1),kopie);		// Punkte lokal berechnen
				return res;
			}
		}
		
		if (Tiefe<=1) {	// Rekursionsende => subTree ist ArrayList von Blttern
			// Brett fr die Bltter Kopieren, Zug durchfhren und Punkte berechnen, nebenbei die Summe aller Punkte bilden
			for (byte i=0; i<len; i++) {
				Zug z = ((Zug)moeglich.get(i));
				float wert = z.BrettLaden(brett);
				z.verwerfeBrett();
				Punkte+=wert;
				if (wert<min) min = wert;
			}
			Punkte /=len;
			Punkte = Punkte*0.5f + min*0.5f;				// Minumum hat Einflu auf die Punkte
			res = len;

		} else { // Pre: this.Pos ist irrelevant

			for (byte i=0; i<len; i++) {
				
				// das Brett fr den jeweiligen Zug berechnen
				Zug z = ((Zug)moeglich.get(i));
				float wert = z.BrettLaden(brett);
				Brett neues_brett = z.verwerfeBrett();		// verwerfen macht unheimlich viel Speicher frei
				
				// dieses Brett rekursiv weiterreichen
				res+=z.Aufbauen((byte)(Tiefe-1),neues_brett);
				
				// Punkte summieren
				Punkte += z.getPunkte()*0.95f+wert*0.05f;		// aktueller Brettwert hat Einflu
			}
			Punkte /=len;
		}


		if (pos.x!=-1) subTree = null; else {			// sofern nicht Wurzel, soviel wie mglich Speicher frei geben
			subTree=moeglich;
			sortSubTree();
		}
		return res;
	}
	
	private void sortSubTree() {
		if (subTree!=null) Collections.sort(subTree);
	}
}


class Brett implements Icon {
	private byte feld[];
	private boolean dran;
	private boolean jump;			// merkt sich, ob der Spieler bersprungen wird
	private byte anzs;
	private byte anzw;

	static final byte weiss = 1;
	static final byte schwarz = -1;
	static final byte frei = 0;
	
	private void Feld(int x, int y, byte was) 	{ feld[x+(y<<3)]=was; }
	private byte Feld(int x, int y)		{ return feld[x+(y<<3)]; }

	Brett() {
		feld = new byte[8*8];
		Feld(3,3,schwarz);
		Feld(4,3,weiss);
		Feld(4,4,schwarz);
		Feld(3,4,weiss);
		anzs=2;
		anzw=2;

		dran=true;
		jump=false;
		statusbyte = 0;
	}

	Brett(Brett b) {
		feld = b.copyFeld();
		dran=b.WerIstDran();
		anzs=b.WievielSchwarze();
		anzw=b.WievielWeisse();
		jump = b.GetJump();
	}
	
	public boolean WerIstDran() 	{ return dran; }
	public byte WievielSchwarze()	{ return anzs; }
	public byte WievielWeisse()	{ return anzw; }
	public byte[] copyFeld() 	{ return (byte[])feld.clone(); }
	public boolean GetJump()	{ return jump; }
	
	public float BrettWert() {					// ermittelt den Punkte-Unterschied
		float delta=anzw-anzs;					// Brettwert > 0 => fr schwarz besser
		
		if (Feld(0,0)==schwarz) delta-=1000; else if (Feld(0,0)==weiss) delta+=1000;
		if (Feld(7,0)==schwarz) delta-=1000; else if (Feld(7,0)==weiss) delta+=1000;
		if (Feld(0,7)==schwarz) delta-=1000; else if (Feld(0,7)==weiss) delta+=1000;
		if (Feld(7,7)==schwarz) delta-=1000; else if (Feld(7,7)==weiss) delta+=1000;
		
		for (byte i=1; i<7; i++) {
			if (Feld(i,0)==schwarz) delta-=50; else if (Feld(i,0)==weiss) delta+=50;
			if (Feld(i,7)==schwarz) delta-=50; else if (Feld(i,7)==weiss) delta+=50;
			if (Feld(0,i)==schwarz) delta-=50; else if (Feld(0,i)==weiss) delta+=50;
			if (Feld(7,i)==schwarz) delta-=50; else if (Feld(7,i)==weiss) delta+=50;
		}
		return delta;
	}
	
	public String toString() {
		char decode[] = {'X','.','O'};
		String res="\n\n  A B C D E F G H     "+anzs+" X  :  "+anzw+" O";
		for (int i=0; i<8; i++) { res+="\n"+(i+1); for (int j=0; j<8; j++) res+=" "+decode[Feld(i,j)+1]; res+=" "+(i+1); }
		res+="\n  A B C D E F G H";
		return res;
	}
	
	private boolean scanline(int x, int y, int dx, int dy, boolean action) {	// scant eine Gerade und ndert ggf das Brett
		// Pre: x,y aus [0..7]
		if (Feld(x,y)!=frei) return false;
		final byte ich=dran?schwarz:weiss;
		final byte Gegner=dran?weiss:schwarz;
		boolean fine=false;

		x+=dx;
		y+=dy;

		while ((x>=0)&&(x<8)&&(y>=0)&&(y<8)) {
			if (Feld(x,y)==Gegner) fine=true; else return fine&&(Feld(x,y)==ich);
			if (action) {
				Feld(x,y,ich);
				if (dran) { anzs++; anzw--;} else { anzw++; anzs--; }
			}
			x+=dx;
			y+=dy;
		}
		
		return false;
	}


	public boolean erlaubt(int x,int y) {		// Pre: x,y aus [0..7]	returns: Richtung, 0 keine
		if (Feld(x,y)!=frei) return false;		// Stelle schon belegt
		boolean res=scanline(x,y,1,0,false);
		res=res?true:scanline(x,y,-1,0,false);	// x->0
		
		res=res?true:scanline(x,y,0,1,false);		// y->7
		res=res?true:scanline(x,y,0,-1,false);	// y->0

		res=res?true:scanline(x,y,1,1,false);		// y->7  x->7
		res=res?true:scanline(x,y,1,-1,false);	// y->0  x->7
		
		res=res?true:scanline(x,y,-1,1,false);	// y->7  x->0
		res=res?true:scanline(x,y,-1,-1,false);	// y->0  x->0
		return res;
	}

	public void moveTo(int x,int y) {			// ndert den Bewertungbaum nicht
		// Pre: erlaubt(x,y) = true
		if (scanline(x,y,1,0,false))	scanline(x,y, 1,0,true);
		if (scanline(x,y,-1,0,false))  scanline(x,y,-1,0,true);
		
		if (scanline(x,y,0, 1,false))  scanline(x,y,0, 1,true);
		if (scanline(x,y,0,-1,false))  scanline(x,y,0,-1,true);

		if (scanline(x,y,1, 1,false))  scanline(x,y,1, 1,true);
		if (scanline(x,y,1,-1,false))  scanline(x,y,1,-1,true);
		
		if (scanline(x,y,-1, 1,false)) scanline(x,y,-1, 1,true);
		if (scanline(x,y,-1,-1,false)) scanline(x,y,-1,-1,true);
		
		Feld(x,y,dran?schwarz:weiss);
		if (dran) anzs++; else anzw++;
		dran=!dran;
	}
	
	public void next_player() { dran=!dran; }


	private boolean TestWin() {		// Testet, ob Spiel Beenden und gibt Sieger aus
		String won=anzs>anzw?"X":(anzs==anzw?"niemand":"O");
		if (anzs+anzw!=64) {
			// alle Felder Prfen, falls noch ein erlaubtes existiert, geht es weiter
			for (int i=0; i<8; i++) for (int j=0; j<8; j++)
				if (erlaubt(i,j)) { jump=false; return false; }
		}
		
		if (!jump) { jump=true; return false; }	// spieler berspringen
		
		Status = ""+won+" hat das Spiel gewonnen.";
		statusbyte = 2;
		return true;
	}

	private void macheZug(int x, int y, boolean art, byte Level) {
		if (!erlaubt(x,y)) { Status = Zug.translate(y,x)+" ist nicht erlaubt."; return; }
		moveTo(x,y);
		statusbyte = 0;
		
		if (TestWin()) return;
		if (jump) { 
			dran=!dran;
			if (TestWin()) return;
			jump=false;
		}
		
		if (!art&&!dran) compZug(Level);
	}
	
	
	public void compZug(byte Level) {		// startet den Computer-Thread
		statusbyte = 1;
		compZugThread th = new compZugThread(this,Level);
		th.start();
	}
	
	public void compZugEnde(int anz, Zug Wurzel, byte Level) {
		Pos p=Wurzel.getBestPos();
		Status = "Ich habe "+anz+" Positionen berechnet und setze nach "+Zug.translate(p.y,p.x)+".";
		moveTo(p.x,p.y);
		statusbyte = 0;

		if (TestWin()) return;
		if (jump) { 
			dran=!dran;
			if (TestWin()) return;
			jump=false;
		}
		
		if (!dran) compZug(Level);
		link.repaint();
	}
		
	public int getIconHeight() 	{ return (TileHoehe<<3)+120; }
	public int getIconWidth()	{ return (TileBreite<<3)+200; }

	private byte statusbyte=0;	// 0, wartet auf Eingabe	
	private String Status="";
	private Surface link = null;

	private static final String stdMsg[] = {"Fehler","Bitte warten, ich berechne noch.","Das Spiel ist bereits beendet.","interner Zug-Fehler"};
	private static final int TileBreite = 50;
	private static final int TileHoehe = 40;
	
	public void paintIcon(Component comp, Graphics g, int xoff, int yoff) {
		Gifs.select[2].paintIcon(comp,g,30,30);			// Background
		
		for (byte i=0; i<8; i++) {
			String s = "" + (i+1);
			g.drawString(s,10,TileHoehe*i+55);
			g.drawString(s,45+(TileBreite<<3),TileHoehe*i+55);
			char c[] = new char[1];
			c[0]=(char)('A'+i);
			s = new String(c);
			g.drawString(s,50+TileBreite*i,20);
			g.drawString(s,50+TileBreite*i,(TileHoehe<<3)+50);
		}
		
		int x=0, y=0;
		for (int n=0; n<64; n++) {
			if (x>=(TileBreite<<3)) {y+=TileHoehe; x=0;}
			g.drawRect(30+x,30+y,TileBreite,TileHoehe);
			switch (feld[n]) {
				case weiss: 	Gifs.select[3].paintIcon(comp,g,31+x,31+y); break;
				case schwarz:	Gifs.select[4].paintIcon(comp,g,31+x,31+y); break;
				case frei:	if (erlaubt(n&7,n>>3)) Gifs.select[5].paintIcon(comp,g,30+x,30+y); break;
			}
			x+=TileBreite;
		}

		g.setFont(new Font(null,Font.BOLD,18));
		g.drawString(Status,30,(TileHoehe<<3)+100);
		if (statusbyte!=2) {
			g.drawString("am Zug:",(TileBreite<<3)+90,60);
			Gifs.select[dran?4:3].paintIcon(comp,g,(TileBreite<<3)+100,80);
		}
		g.drawString("Punkte:",(TileBreite<<3)+90,100+(TileHoehe<<1));
		g.drawString(""+WievielSchwarze(),(TileBreite<<3)+90,150+(TileHoehe<<1));
		Gifs.select[4].paintIcon(comp,g,(TileBreite<<3)+115,120+(TileHoehe<<1));
		g.drawString(""+WievielWeisse(),(TileBreite<<3)+90,150+(TileHoehe<<2));
		Gifs.select[3].paintIcon(comp,g,(TileBreite<<3)+115,120+(TileHoehe<<2));
		
	}
	
	public void MouseClick(Surface l, int x, int y, boolean art, byte level) {
		if (statusbyte!=0) {
			Status = stdMsg[statusbyte];
			link.repaint();
			return;
		}
		
		if (!dran&&!art) {
			statusbyte=3;
			Status = stdMsg[statusbyte];
			link.repaint();
			return;
		}
		
		x-=30; y-=30;
		if ((x<0)||(y<0)) return;
		x /= TileBreite;
		y /= TileHoehe;
		if ((x>=8)||(y>=8)) return;
		Status = "Nach "+Zug.translate(y,x)+" gesetzt.";
		link = l;
		macheZug(x,y,art,level);
		link.repaint();
	}
}

class compZugThread extends Thread {
	private Brett link;
	private byte Level;
	
	compZugThread( Brett l, byte Tiefe) {
		link = l;
		Level = Tiefe;
	}
	
	public void run() {
		Zug Wurzel = new Zug(-1,-1);
		int anz = Wurzel.Aufbauen((byte)((Level<<1)-1),link);				// Stellt die Baumeigenschaften her
		link.compZugEnde(anz,Wurzel,Level);
	}
}

// -------------------------------------------------------------------- Dialog-Klassen

class Gifs extends Thread {
	
	public static final ImageIcon select[] = {
		new ImageIcon("bob.gif"),
		new ImageIcon("pc.gif"),
		new ImageIcon("back.gif"),
		new ImageIcon("O.gif"),
		new ImageIcon("X.gif"),
		new ImageIcon("able.gif")
	};
	
	public void start() { run(); }	// sobald ein Objekt entsteht starten
	
	public void run() {			// Thread arbeitet im Hintergrund und soll schonmal die Bilder laden
		for (int i=0; i<select.length; i++) select[i].getImage();
	}
}

class Press extends JButton {
	
	public Press(String s) {
		super(s);
		setFocusPainted(false);
	}
	
	public Press(String s, String h) {
		this(s);
		setToolTipText(h);
	}

	public Press(String s, String h, ActionListener al) {
		this(s,h);
		addActionListener(al);
	}

}

class Surface extends JPanel {
	private Brett data;
	private JLabel label;

	public Surface(final boolean type, final byte level) {
		
		final Surface self = this;

		MouseAdapter ma = new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				Rectangle r = label.getBounds(null);
				e.translatePoint(-(int)r.getMinX(),-(int)r.getMinY());
				data.MouseClick(self, e.getX(), e.getY(), type, level);
			}
		};
		
		data = new Brett();
		label = new JLabel(data);
		add(label);
		addMouseListener(ma);
	}
}

class Dialog extends JFrame {
	private JRadioButton rb1;
	private JSlider sl;
	
	public void display()	{	// packen, zentrieren und darstellen
		pack();
		Dimension d = Toolkit.getDefaultToolkit().getScreenSize();		
		setLocation((d.width-getSize().width)>>1,(d.height-getSize().height)>>1);
		show();
		toFront();
		toFront();
	}
	
	public void startgame(boolean type, byte level) {
		Surface surface = new Surface(type, level);
		getContentPane().add(surface);
		setResizable(true);
		display();
	}
	
	public void start() {
		
		ActionListener al_exit = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				System.exit(0);
			}
		};
		
		ActionListener al_next = new ActionListener() {
			
			public void actionPerformed( ActionEvent e ) {
				boolean type = rb1.isSelected();
				rb1=null;
				hide();
				getContentPane().removeAll();
				
				ActionListener al2_next = new ActionListener() {
					public void actionPerformed( ActionEvent e) {
						byte level = (byte)sl.getValue();
						sl = null;
						hide();
						getContentPane().removeAll();
						startgame(false,level);
					}
				};
				
				if (type) startgame(true,(byte)0); else {	// Schwierigkeitsgrad erfragen
					JPanel panel = new JPanel(true);
					panel.setLayout( new GridLayout(3,1));
					panel.add(new JLabel("  Whlen Sie den Schwierigkeitsgrad.  "));
					
					sl = new JSlider(JSlider.HORIZONTAL,1,4,2);
					sl.setMinorTickSpacing(1);
					sl.setPaintLabels(true);
					sl.setPaintTicks(true);
					
					sl.setLabelTable(sl.createStandardLabels(1,1));
					sl.setSnapToTicks(true);
					panel.add(sl);
					
					panel.add(new Press("weiter","jetzt geht's los",al2_next));
					getContentPane().add(panel);
					display();
				}
			}
		};
	
		hide();
		getContentPane().removeAll();
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setResizable(false);
		setTitle("Reversi 2.0  - Freeware von www.drgames.de");
		
		JPanel panel = new JPanel(true);
		panel.setLayout( new GridLayout(4,2) );
		panel.add(new JLabel("  Whlen Sie die Spielart."));
		panel.add(new JLabel(" "));
		
		panel.add(new JLabel(Gifs.select[0]));
		panel.add(new JLabel(Gifs.select[1]));
		
		rb1 = new JRadioButton( "Mensch : Mensch",false);
		rb1.setFocusPainted(false);
		
		JRadioButton rb2 = new JRadioButton( "Mensch : Computer" ,true);
		rb2.setFocusPainted(false);
		ButtonGroup g = new ButtonGroup();
		g.add( rb1 ); g.add( rb2 );
		panel.add(rb1);
		panel.add(rb2);

		panel.add(new Press("abbrechen","Spiel verlassen",al_exit));
		panel.add(new Press("weiter","Spiel starten",al_next));
		getContentPane().add(panel);
		display();
	}
}

public class reversi2 {

	public static void main(String args[]) {
		Gifs pre_load = new Gifs();
		Dialog dlg = new Dialog();
		dlg.start();
	}
	
}

